# Copyright (c) 2020 Nekokatt
# Copyright (c) 2021-present davfsa
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import annotations

import mock
import pytest

from hikari import channels, urls
from hikari import undefined
from hikari import webhooks
from tests.hikari import hikari_test_helpers


class TestExecutableWebhook:
    @pytest.fixture
    def executable_webhook(self):
        return hikari_test_helpers.mock_class_namespace(
            webhooks.ExecutableWebhook, slots_=False, app=mock.AsyncMock()
        )()

    @pytest.mark.asyncio
    async def test_execute_when_no_token(self, executable_webhook):
        executable_webhook.token = None

        with pytest.raises(ValueError, match=r"Cannot send a message using a webhook where we don't know the token"):
            await executable_webhook.execute()

    @pytest.mark.asyncio
    async def test_execute_with_optionals(self, executable_webhook):
        mock_attachment_1 = object()
        mock_attachment_2 = object()
        mock_component = object()
        mock_components = object(), object()
        mock_embed = object()
        mock_embeds = object(), object()
        mock_poll = object()

        result = await executable_webhook.execute(
            content="coooo",
            username="oopp",
            avatar_url="urlurlurl",
            tts=True,
            attachment=mock_attachment_1,
            attachments=mock_attachment_2,
            component=mock_component,
            components=mock_components,
            embed=mock_embed,
            embeds=mock_embeds,
            poll=mock_poll,
            mentions_everyone=False,
            user_mentions=[1235432],
            role_mentions=[65234123],
            flags=64,
        )

        assert result is executable_webhook.app.rest.execute_webhook.return_value
        executable_webhook.app.rest.execute_webhook.assert_awaited_once_with(
            webhook=executable_webhook.webhook_id,
            token=executable_webhook.token,
            content="coooo",
            username="oopp",
            avatar_url="urlurlurl",
            tts=True,
            attachment=mock_attachment_1,
            attachments=mock_attachment_2,
            component=mock_component,
            components=mock_components,
            embed=mock_embed,
            embeds=mock_embeds,
            poll=mock_poll,
            mentions_everyone=False,
            user_mentions=[1235432],
            role_mentions=[65234123],
            flags=64,
        )

    @pytest.mark.asyncio
    async def test_execute_without_optionals(self, executable_webhook):
        result = await executable_webhook.execute()

        assert result is executable_webhook.app.rest.execute_webhook.return_value
        executable_webhook.app.rest.execute_webhook.assert_awaited_once_with(
            webhook=executable_webhook.webhook_id,
            token=executable_webhook.token,
            content=undefined.UNDEFINED,
            username=undefined.UNDEFINED,
            avatar_url=undefined.UNDEFINED,
            tts=undefined.UNDEFINED,
            attachment=undefined.UNDEFINED,
            attachments=undefined.UNDEFINED,
            component=undefined.UNDEFINED,
            components=undefined.UNDEFINED,
            embed=undefined.UNDEFINED,
            embeds=undefined.UNDEFINED,
            poll=undefined.UNDEFINED,
            mentions_everyone=undefined.UNDEFINED,
            user_mentions=undefined.UNDEFINED,
            role_mentions=undefined.UNDEFINED,
            flags=undefined.UNDEFINED,
        )

    @pytest.mark.asyncio
    async def test_fetch_message(self, executable_webhook):
        message = object()
        returned_message = object()
        executable_webhook.app.rest.fetch_webhook_message = mock.AsyncMock(return_value=returned_message)

        returned = await executable_webhook.fetch_message(message)

        assert returned is returned_message

        executable_webhook.app.rest.fetch_webhook_message.assert_awaited_once_with(
            executable_webhook.webhook_id, token=executable_webhook.token, message=message
        )

    @pytest.mark.asyncio
    async def test_fetch_message_when_no_token(self, executable_webhook):
        executable_webhook.token = None
        with pytest.raises(ValueError, match=r"Cannot fetch a message using a webhook where we don't know the token"):
            await executable_webhook.fetch_message(987)

    @pytest.mark.asyncio
    async def test_edit_message(self, executable_webhook):
        message = object()
        embed = object()
        attachment = object()
        component = object()
        components = object()

        returned = await executable_webhook.edit_message(
            message,
            content="test",
            embed=embed,
            embeds=[embed, embed],
            attachment=attachment,
            attachments=[attachment, attachment],
            component=component,
            components=components,
            mentions_everyone=False,
            user_mentions=True,
            role_mentions=[567, 890],
        )

        assert returned is executable_webhook.app.rest.edit_webhook_message.return_value

        executable_webhook.app.rest.edit_webhook_message.assert_awaited_once_with(
            executable_webhook.webhook_id,
            token=executable_webhook.token,
            message=message,
            content="test",
            embed=embed,
            embeds=[embed, embed],
            attachment=attachment,
            attachments=[attachment, attachment],
            component=component,
            components=components,
            mentions_everyone=False,
            user_mentions=True,
            role_mentions=[567, 890],
        )

    @pytest.mark.asyncio
    async def test_edit_message_when_no_token(self, executable_webhook):
        executable_webhook.token = None
        with pytest.raises(ValueError, match=r"Cannot edit a message using a webhook where we don't know the token"):
            await executable_webhook.edit_message(987)

    @pytest.mark.asyncio
    async def test_delete_message(self, executable_webhook):
        message = object()

        await executable_webhook.delete_message(message)

        executable_webhook.app.rest.delete_webhook_message.assert_awaited_once_with(
            executable_webhook.webhook_id, token=executable_webhook.token, message=message
        )

    @pytest.mark.asyncio
    async def test_delete_message_when_no_token(self, executable_webhook):
        executable_webhook.token = None
        with pytest.raises(ValueError, match=r"Cannot delete a message using a webhook where we don't know the token"):
            assert await executable_webhook.delete_message(987)


class TestPartialWebhook:
    @pytest.fixture
    def webhook(self):
        return webhooks.PartialWebhook(
            app=mock.Mock(rest=mock.AsyncMock()),
            id=987654321,
            type=webhooks.WebhookType.CHANNEL_FOLLOWER,
            name="not a webhook",
            avatar_hash="hook",
            application_id=None,
        )

    def test_str(self, webhook):
        assert str(webhook) == "not a webhook"

    def test_str_when_name_is_None(self, webhook):
        webhook.name = None
        assert str(webhook) == "Unnamed webhook ID 987654321"

    def test_mention_property(self, webhook):
        assert webhook.mention == "<@987654321>"

    def test_make_avatar_url_format_set_to_deprecated_ext_argument_if_provided(self, webhook):
        with mock.patch.object(
            channels.routes, "CDN_USER_AVATAR", new=mock.Mock(compile_to_file=mock.Mock(return_value="file"))
        ) as route:
            assert webhook.make_avatar_url(ext="JPEG") == "file"

        route.compile_to_file.assert_called_once_with(
            urls.CDN_URL, user_id=987654321, hash="hook", size=4096, file_format="JPEG", lossless=True
        )

    def test_avatar_url_property(self, webhook):
        assert webhook.avatar_url == webhook.make_avatar_url()

    def test_default_avatar_url(self, webhook):
        assert webhook.default_avatar_url.url == "https://cdn.discordapp.com/embed/avatars/0.png"

    def test_make_avatar_url(self, webhook):
        result = webhook.make_avatar_url(ext="jpeg", size=2048)

        assert result.url == "https://cdn.discordapp.com/avatars/987654321/hook.jpeg?size=2048"

    def test_make_avatar_url_when_no_avatar(self, webhook):
        webhook.avatar_hash = None

        assert webhook.make_avatar_url() is None


class TestIncomingWebhook:
    @pytest.fixture
    def webhook(self):
        return webhooks.IncomingWebhook(
            app=mock.Mock(rest=mock.AsyncMock()),
            id=987654321,
            type=webhooks.WebhookType.CHANNEL_FOLLOWER,
            guild_id=123,
            channel_id=456,
            author=None,
            name="not a webhook",
            avatar_hash=None,
            token="abc123bca",
            application_id=None,
        )

    def test_webhook_id_property(self, webhook):
        assert webhook.webhook_id is webhook.id

    @pytest.mark.asyncio
    async def test_delete(self, webhook):
        webhook.token = None

        await webhook.delete()

        webhook.app.rest.delete_webhook.assert_awaited_once_with(987654321, token=undefined.UNDEFINED)

    @pytest.mark.asyncio
    async def test_delete_uses_token_property(self, webhook):
        webhook.token = "123321"

        await webhook.delete()

        webhook.app.rest.delete_webhook.assert_awaited_once_with(987654321, token="123321")

    @pytest.mark.asyncio
    async def test_delete_use_token_is_true(self, webhook):
        webhook.token = "322312"

        await webhook.delete(use_token=True)

        webhook.app.rest.delete_webhook.assert_awaited_once_with(987654321, token="322312")

    @pytest.mark.asyncio
    async def test_delete_use_token_is_true_without_token(self, webhook):
        webhook.token = None

        with pytest.raises(ValueError, match="This webhook's token is unknown, so cannot be used"):
            await webhook.delete(use_token=True)

        webhook.app.rest.delete_webhook.assert_not_called()

    @pytest.mark.asyncio
    async def test_delete_use_token_is_false(self, webhook):
        webhook.token = "322312"

        await webhook.delete(use_token=False)

        webhook.app.rest.delete_webhook.assert_awaited_once_with(987654321, token=undefined.UNDEFINED)

    @pytest.mark.asyncio
    async def test_edit(self, webhook):
        webhook.token = None
        webhook.app.rest.edit_webhook.return_value = mock.Mock(webhooks.IncomingWebhook)
        mock_avatar = object()

        result = await webhook.edit(name="OK", avatar=mock_avatar, channel=33333, reason="byebye")

        assert result is webhook.app.rest.edit_webhook.return_value
        webhook.app.rest.edit_webhook.assert_awaited_once_with(
            987654321, token=undefined.UNDEFINED, name="OK", avatar=mock_avatar, channel=33333, reason="byebye"
        )

    @pytest.mark.asyncio
    async def test_edit_uses_token_property(self, webhook):
        webhook.token = "aye"
        webhook.app.rest.edit_webhook.return_value = mock.Mock(webhooks.IncomingWebhook)
        mock_avatar = object()

        result = await webhook.edit(name="bye", avatar=mock_avatar, channel=33333, reason="byebye")

        assert result is webhook.app.rest.edit_webhook.return_value
        webhook.app.rest.edit_webhook.assert_awaited_once_with(
            987654321, token="aye", name="bye", avatar=mock_avatar, channel=33333, reason="byebye"
        )

    @pytest.mark.asyncio
    async def test_edit_when_use_token_is_true(self, webhook):
        webhook.token = "owoowow"
        webhook.app.rest.edit_webhook.return_value = mock.Mock(webhooks.IncomingWebhook)
        mock_avatar = object()

        result = await webhook.edit(use_token=True, name="hiu", avatar=mock_avatar, channel=231, reason="sus")

        assert result is webhook.app.rest.edit_webhook.return_value
        webhook.app.rest.edit_webhook.assert_awaited_once_with(
            987654321, token="owoowow", name="hiu", avatar=mock_avatar, channel=231, reason="sus"
        )

    @pytest.mark.asyncio
    async def test_edit_when_use_token_is_true_and_no_token(self, webhook):
        webhook.token = None

        with pytest.raises(ValueError, match="This webhook's token is unknown, so cannot be used"):
            await webhook.edit(use_token=True)

        webhook.app.rest.edit_webhook.assert_not_called()

    @pytest.mark.asyncio
    async def test_edit_when_use_token_is_false(self, webhook):
        webhook.token = "owoowow"
        webhook.app.rest.edit_webhook.return_value = mock.Mock(webhooks.IncomingWebhook)
        mock_avatar = object()

        result = await webhook.edit(use_token=False, name="eee", avatar=mock_avatar, channel=231, reason="rrr")

        assert result is webhook.app.rest.edit_webhook.return_value
        webhook.app.rest.edit_webhook.assert_awaited_once_with(
            987654321, token=undefined.UNDEFINED, name="eee", avatar=mock_avatar, channel=231, reason="rrr"
        )

    @pytest.mark.asyncio
    async def test_fetch_channel(self, webhook):
        webhook.app.rest.fetch_channel.return_value = mock.Mock(channels.GuildTextChannel)

        assert await webhook.fetch_channel() is webhook.app.rest.fetch_channel.return_value

        webhook.app.rest.fetch_channel.assert_awaited_once_with(webhook.channel_id)

    @pytest.mark.asyncio
    async def test_fetch_self(self, webhook):
        webhook.token = None
        webhook.app.rest.fetch_webhook.return_value = mock.Mock(webhooks.IncomingWebhook)

        result = await webhook.fetch_self()

        assert result is webhook.app.rest.fetch_webhook.return_value
        webhook.app.rest.fetch_webhook.assert_awaited_once_with(987654321, token=undefined.UNDEFINED)

    @pytest.mark.asyncio
    async def test_fetch_self_uses_token_property(self, webhook):
        webhook.token = "no gnomo"
        webhook.app.rest.fetch_webhook.return_value = mock.Mock(webhooks.IncomingWebhook)

        result = await webhook.fetch_self()

        assert result is webhook.app.rest.fetch_webhook.return_value
        webhook.app.rest.fetch_webhook.assert_awaited_once_with(987654321, token="no gnomo")

    @pytest.mark.asyncio
    async def test_fetch_self_when_use_token_is_true(self, webhook):
        webhook.token = "no momo"
        webhook.app.rest.fetch_webhook.return_value = mock.Mock(webhooks.IncomingWebhook)

        result = await webhook.fetch_self(use_token=True)

        assert result is webhook.app.rest.fetch_webhook.return_value
        webhook.app.rest.fetch_webhook.assert_awaited_once_with(987654321, token="no momo")

    @pytest.mark.asyncio
    async def test_fetch_self_when_use_token_is_true_without_token_property(self, webhook):
        webhook.token = None

        with pytest.raises(ValueError, match="This webhook's token is unknown, so cannot be used"):
            await webhook.fetch_self(use_token=True)

        webhook.app.rest.fetch_webhook.assert_not_called()

    @pytest.mark.asyncio
    async def test_fetch_self_when_use_token_is_false(self, webhook):
        webhook.token = "no momo"
        webhook.app.rest.fetch_webhook.return_value = mock.Mock(webhooks.IncomingWebhook)

        result = await webhook.fetch_self(use_token=False)

        assert result is webhook.app.rest.fetch_webhook.return_value
        webhook.app.rest.fetch_webhook.assert_awaited_once_with(987654321, token=undefined.UNDEFINED)


class TestChannelFollowerWebhook:
    @pytest.fixture
    def webhook(self):
        return webhooks.ChannelFollowerWebhook(
            app=mock.Mock(rest=mock.AsyncMock()),
            id=987654321,
            type=webhooks.WebhookType.CHANNEL_FOLLOWER,
            guild_id=123,
            channel_id=456,
            author=None,
            name="not a webhook",
            avatar_hash=None,
            application_id=None,
            source_channel=object(),
            source_guild=object(),
        )

    @pytest.mark.asyncio
    async def test_delete(self, webhook):
        await webhook.delete()

        webhook.app.rest.delete_webhook.assert_awaited_once_with(987654321)

    @pytest.mark.asyncio
    async def test_edit(self, webhook):
        mock_avatar = object()
        webhook.app.rest.edit_webhook.return_value = mock.Mock(webhooks.ChannelFollowerWebhook)

        result = await webhook.edit(name="hi", avatar=mock_avatar, channel=43123, reason="ok")

        assert result is webhook.app.rest.edit_webhook.return_value
        webhook.app.rest.edit_webhook.assert_awaited_once_with(
            987654321, name="hi", avatar=mock_avatar, channel=43123, reason="ok"
        )

    @pytest.mark.asyncio
    async def test_fetch_channel(self, webhook):
        webhook.app.rest.fetch_channel.return_value = mock.Mock(channels.GuildTextChannel)

        assert await webhook.fetch_channel() is webhook.app.rest.fetch_channel.return_value

        webhook.app.rest.fetch_channel.assert_awaited_once_with(webhook.channel_id)

    @pytest.mark.asyncio
    async def test_fetch_self(self, webhook):
        webhook.app.rest.fetch_webhook.return_value = mock.Mock(webhooks.ChannelFollowerWebhook)

        result = await webhook.fetch_self()

        assert result is webhook.app.rest.fetch_webhook.return_value
        webhook.app.rest.fetch_webhook.assert_awaited_once_with(987654321)



# Copyright (c) 2020 Nekokatt
# Copyright (c) 2021-present davfsa
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Models and enums used for Discord's Components interaction flow."""

from __future__ import annotations

__all__: typing.Sequence[str] = (
    "COMPONENT_RESPONSE_TYPES",
    "ComponentInteraction",
    "ComponentInteractionMetadata",
    "ComponentResponseTypesT",
)

import typing

import attrs

from hikari.interactions import base_interactions

if typing.TYPE_CHECKING:
    from hikari import components as components_
    from hikari import messages
    from hikari import snowflakes
    from hikari.api import special_endpoints


_DEFERRED_TYPES: typing.AbstractSet[_DeferredTypesT] = frozenset(
    [base_interactions.ResponseType.DEFERRED_MESSAGE_CREATE, base_interactions.ResponseType.DEFERRED_MESSAGE_UPDATE]
)
_DeferredTypesT = typing.Literal[
    base_interactions.ResponseType.DEFERRED_MESSAGE_CREATE, 5, base_interactions.ResponseType.DEFERRED_MESSAGE_UPDATE, 6
]
_IMMEDIATE_TYPES: typing.AbstractSet[_ImmediateTypesT] = frozenset(
    [base_interactions.ResponseType.MESSAGE_CREATE, base_interactions.ResponseType.MESSAGE_UPDATE]
)
_ImmediateTypesT = typing.Literal[
    base_interactions.ResponseType.MESSAGE_CREATE, 4, base_interactions.ResponseType.MESSAGE_UPDATE, 7
]


COMPONENT_RESPONSE_TYPES: typing.Final[typing.AbstractSet[ComponentResponseTypesT]] = frozenset(
    [*_DEFERRED_TYPES, *_IMMEDIATE_TYPES]
)
"""Set of the response types which are valid for a component interaction.

This includes:

* [`hikari.interactions.base_interactions.ResponseType.MESSAGE_CREATE`][]
* [`hikari.interactions.base_interactions.ResponseType.DEFERRED_MESSAGE_CREATE`][]
* [`hikari.interactions.base_interactions.ResponseType.DEFERRED_MESSAGE_UPDATE`][]
* [`hikari.interactions.base_interactions.ResponseType.MESSAGE_UPDATE`][]
"""

ComponentResponseTypesT = typing.Union[_ImmediateTypesT, _DeferredTypesT]
"""Type-hint of the response types which are valid for a component interaction.

The following types are valid for this:

* [`hikari.interactions.base_interactions.ResponseType.MESSAGE_CREATE`][]/`4`
* [`hikari.interactions.base_interactions.ResponseType.DEFERRED_MESSAGE_CREATE`][]/`5`
* [`hikari.interactions.base_interactions.ResponseType.DEFERRED_MESSAGE_UPDATE`][]/`6`
* [`hikari.interactions.base_interactions.ResponseType.MESSAGE_UPDATE`][]/`7`
"""


@attrs.define(unsafe_hash=True, weakref_slot=False)
class ComponentInteraction(
    base_interactions.MessageResponseMixin[ComponentResponseTypesT],
    base_interactions.ModalResponseMixin,
    base_interactions.PremiumResponseMixin,
):
    """Represents a component interaction on Discord."""

    component_type: components_.ComponentType | int = attrs.field(eq=False)
    """The type of component which triggers this interaction.

    !!! note
        This will never be [`hikari.components.ButtonStyle.LINK`][] as link buttons don't trigger
        interactions.
    """

    custom_id: str = attrs.field(eq=False)
    """Developer defined ID of the component which triggered this interaction."""

    values: typing.Sequence[str] = attrs.field(eq=False)
    """Sequence of the values which were selected for a select menu component."""

    resolved: base_interactions.ResolvedOptionData | None = attrs.field(eq=False, hash=False, repr=False)
    """Mappings of the objects resolved for the provided command options."""

    message: messages.Message = attrs.field(eq=False, repr=False)
    """Object of the message the components for this interaction are attached to."""

    def build_response(self, type_: _ImmediateTypesT, /) -> special_endpoints.InteractionMessageBuilder:
        """Get a message response builder for use in the REST server flow.

        !!! note
            For interactions received over the gateway
            [`hikari.interactions.component_interactions.ComponentInteraction.create_initial_response`][]
            should be used to set the interaction response message.

        Parameters
        ----------
        type_
            The type of immediate response this should be.

            This may be one of the following:

            * [`hikari.interactions.base_interactions.ResponseType.MESSAGE_CREATE`][]
            * [`hikari.interactions.base_interactions.ResponseType.MESSAGE_UPDATE`][]

        Examples
        --------
        ```py
        async def handle_component_interaction(
            interaction: ComponentInteraction,
        ) -> InteractionMessageBuilder:
            return (
                interaction.build_response(ResponseType.MESSAGE_UPDATE)
                .add_embed(Embed(description="Hi there"))
                .set_content("Konnichiwa")
            )
        ```

        Returns
        -------
        hikari.api.special_endpoints.InteractionMessageBuilder
            Interaction message response builder object.
        """
        if type_ not in _IMMEDIATE_TYPES:
            msg = "Invalid type passed for an immediate response"
            raise ValueError(msg)

        return self.app.rest.interaction_message_builder(type_)

    def build_deferred_response(self, type_: _DeferredTypesT, /) -> special_endpoints.InteractionDeferredBuilder:
        """Get a deferred message response builder for use in the REST server flow.

        !!! note
            For interactions received over the gateway
            [`hikari.interactions.component_interactions.ComponentInteraction.create_initial_response`][]
            should be used to set the interaction response message.

        !!! note
            Unlike [`hikari.api.special_endpoints.InteractionMessageBuilder`][],
            the result of this call can be returned as is without any modifications
            being made to it.

        Parameters
        ----------
        type_
            The type of deferred response this should be.

            This may be one of the following:

            * [`hikari.interactions.base_interactions.ResponseType.DEFERRED_MESSAGE_CREATE`][]
            * [`hikari.interactions.base_interactions.ResponseType.DEFERRED_MESSAGE_UPDATE`][]

        Returns
        -------
        hikari.api.special_endpoints.InteractionDeferredBuilder
            Deferred interaction message response builder object.
        """
        if type_ not in _DEFERRED_TYPES:
            msg = "Invalid type passed for a deferred response"
            raise ValueError(msg)

        return self.app.rest.interaction_deferred_builder(type_)


@attrs.define(unsafe_hash=True, kw_only=True, weakref_slot=False)
class ComponentInteractionMetadata(base_interactions.PartialInteractionMetadata):
    """The interaction metadata for a component belonging to a message."""

    original_response_message_id: snowflakes.Snowflake | None = attrs.field(eq=False, hash=False, repr=True)
    """The ID of the original response message, present only on follow-up messages."""

    interacted_message_id: snowflakes.Snowflake = attrs.field(eq=False, hash=False, repr=True)
    """The ID of the message that contained the interactive component"""



# Copyright (c) 2020 Nekokatt
# Copyright (c) 2021-present davfsa
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import annotations

import builtins
import operator
import warnings

import mock
import pytest

from hikari.internal import deprecation
from hikari.internal import enums


class TestEnum:
    @mock.patch.object(enums, "_Enum", new=NotImplemented)
    def test_init_first_enum_type_populates_Enum(self):
        class Enum(metaclass=enums._EnumMeta):
            pass

        assert enums._Enum is Enum

    @mock.patch.object(enums, "_Enum", new=NotImplemented)
    def test_init_first_enum_type_with_wrong_name_and_no_bases_raises_TypeError(self):
        with pytest.raises(TypeError):

            class Potato(metaclass=enums._EnumMeta):
                pass

        assert enums._Enum is NotImplemented

    def test_init_second_enum_type_with_no_bases_does_not_change_Enum_attribute_and_raises_TypeError(self):
        expect = enums._Enum

        with pytest.raises(TypeError):

            class Enum(metaclass=enums._EnumMeta):
                pass

        assert enums._Enum is expect

    @pytest.mark.parametrize(
        ("args", "kwargs"),
        [([str], {"metaclass": enums._EnumMeta}), ([enums.Enum], {"metaclass": enums._EnumMeta}), ([enums.Enum], {})],
    )
    def test_init_enum_type_with_one_base_is_TypeError(self, args, kwargs):
        with pytest.raises(TypeError):

            class Enum(*args, **kwargs):
                pass

    @pytest.mark.parametrize(
        ("args", "kwargs"), [([enums.Enum, str], {"metaclass": enums._EnumMeta}), ([enums.Enum, str], {})]
    )
    def test_init_enum_type_with_bases_in_wrong_order_is_TypeError(self, args, kwargs):
        with pytest.raises(TypeError):

            class Enum(*args, **kwargs):
                pass

    def test_init_with_more_than_2_types(self):
        with pytest.raises(TypeError):

            class Enum(enums.Enum, str, int):
                pass

    def test_init_with_less_than_2_types(self):
        with pytest.raises(TypeError):

            class Enum(enums.Enum):
                pass

    def test_init_enum_type_default_docstring_set(self):
        class Enum(str, enums.Enum):
            pass

        assert Enum.__doc__ == "An enumeration."

    def test_init_enum_type_disallows_objects_that_are_not_instances_of_the_first_base(self):
        with pytest.raises(TypeError):

            class Enum(str, enums.Enum):
                foo = 1

    def test_init_enum_type_allows_any_object_if_it_has_a_dunder_name(self):
        class Enum(str, enums.Enum):
            __foo__ = 1
            __bar = 2

        assert Enum is not None

    def test_init_enum_type_allows_any_object_if_it_has_a_sunder_name(self):
        class Enum(str, enums.Enum):
            _foo_ = 1
            _bar = 2

        assert Enum is not None

    def test_init_enum_type_allows_methods(self):
        class Enum(int, enums.Enum):
            def foo(self):
                return "foo"

        assert Enum.foo(12) == "foo"

    def test_init_enum_type_allows_classmethods(self):
        class Enum(int, enums.Enum):
            @classmethod
            def foo(cls):
                assert cls is Enum
                return "foo"

        assert Enum.foo() == "foo"

    def test_init_enum_type_allows_staticmethods(self):
        class Enum(int, enums.Enum):
            @staticmethod
            def foo():
                return "foo"

        assert Enum.foo() == "foo"

    def test_init_enum_type_allows_descriptors(self):
        class Enum(int, enums.Enum):
            @property
            def foo(self):
                return "foo"

        assert isinstance(Enum.foo, property)

    def test_init_enum_type_maps_names_in_members(self):
        class Enum(int, enums.Enum):
            foo = 9
            bar = 18
            baz = 27

            @staticmethod
            def sm():
                pass

            @classmethod
            def cm(cls):
                pass

            def m(self):
                pass

            @property
            def p(self):
                pass

            __dunder__ = "aaa"
            _sunder_ = "bbb"
            __priv = "ccc"
            _prot = "ddd"

        assert Enum.__members__ == {"foo": 9, "bar": 18, "baz": 27}

    def test_init_with_invalid_name(self):
        with pytest.raises(TypeError):

            class Enum(int, enums.Enum):
                mro = 420

    def test_init_with_unhashable_value(self):
        with mock.patch.object(builtins, "hash", side_effect=TypeError):
            with pytest.raises(TypeError):

                class Enum(int, enums.Enum):
                    test = 420

    def test_init_with_duplicate(self):
        with pytest.raises(TypeError):

            class Enum(int, enums.Enum):
                test = 123
                test = 321

    def test_call_when_member(self):
        class Enum(int, enums.Enum):
            foo = 9
            bar = 18
            baz = 27

        returned = Enum(9)
        assert returned == Enum.foo
        assert type(returned) is Enum

    def test_call_when_not_member(self):
        class Enum(int, enums.Enum):
            foo = 9
            bar = 18
            baz = 27

        returned = Enum(69)
        assert returned == 69
        assert type(returned) is not Enum

    def test_getitem(self):
        class Enum(int, enums.Enum):
            foo = 9
            bar = 18
            baz = 27

        returned = Enum["foo"]
        assert returned == Enum.foo
        assert type(returned) is Enum

    def test_contains(self):
        class Enum(int, enums.Enum):
            foo = 9
            bar = 18
            baz = 27

        assert 9 in Enum
        assert 100 not in Enum

    def test_name(self):
        class Enum(int, enums.Enum):
            foo = 9
            bar = 18
            baz = 27

        assert Enum.foo.name == "foo"

    def test_iter(self):
        class Enum(int, enums.Enum):
            foo = 9
            bar = 18
            baz = 27

        a = []
        for i in Enum:
            a.append(i)

        assert a == [Enum.foo, Enum.bar, Enum.baz]

    def test_repr(self):
        class Enum(int, enums.Enum):
            foo = 9
            bar = 18
            baz = 27

        assert repr(Enum) == "<enum Enum>"
        assert repr(Enum.foo) == "<Enum.foo: 9>"

    def test_str(self):
        class Enum(int, enums.Enum):
            foo = 9
            bar = 18
            baz = 27

        assert str(Enum) == "<enum Enum>"
        assert str(Enum.foo) == "foo"

    def test_can_overwrite_method(self):
        class TestEnum1(str, enums.Enum):
            FOO = "foo"

            def __str__(self) -> str:
                return "Ok"

        assert str(TestEnum1.FOO) == "Ok"

    @pytest.mark.parametrize(("type_", "value"), [(int, 42), (str, "ok"), (bytes, b"no"), (float, 4.56), (complex, 3j)])
    def test_inherits_type_dunder_method_behaviour(self, type_, value):
        class TestEnum(type_, enums.Enum):
            BAR = value

        result = type_(TestEnum.BAR)

        assert type(result) is type_
        assert result == value

    def test_allows_overriding_methods(self):
        class TestEnum(int, enums.Enum):
            BAR = 2222

            def __int__(self):
                return 53

        assert int(TestEnum.BAR) == 53


class TestIntFlag:
    @mock.patch.object(enums, "_Flag", new=NotImplemented)
    def test_init_first_flag_type_populates_Flag(self):
        class Flag(metaclass=enums._FlagMeta):
            a = 1

        assert enums._Flag is Flag

    @mock.patch.object(enums, "_Flag", new=NotImplemented)
    def test_init_first_flag_type_with_wrong_name_and_no_bases_raises_TypeError(self):
        with pytest.raises(TypeError):

            class Potato(metaclass=enums._FlagMeta):
                a = 1

        assert enums._Flag is NotImplemented

    def test_init_second_flag_type_with_no_bases_does_not_change_Flag_attribute_and_raises_TypeError(self):
        expect = enums._Flag

        with pytest.raises(TypeError):

            class Flag(metaclass=enums._FlagMeta):
                a = 1

        assert enums._Flag is expect

    def test_init_flag_type_default_docstring_set(self):
        class Flag(enums.Flag):
            a = 1

        assert Flag.__doc__ == "An enumeration."

    def test_init_flag_type_disallows_objects_that_are_not_instances_int(self):
        with pytest.raises(TypeError):

            class Flag(enums.Flag):
                a = 1
                foo = "hi"

    def test_init_flag_type_disallows_other_bases(self):
        with pytest.raises(TypeError):

            class Flag(float, enums.Flag):
                a = 1

    def test_init_flag_type_allows_any_object_if_it_has_a_dunder_name(self):
        class Flag(enums.Flag):
            __foo__ = 1
            __bar = 2
            a = 3

        assert Flag is not None

    def test_init_flag_type_allows_any_object_if_it_has_a_sunder_name(self):
        class Flag(enums.Flag):
            _foo_ = 1
            _bar = 2
            a = 3

        assert Flag is not None

    def test_init_flag_type_allows_methods(self):
        class Flag(enums.Flag):
            A = 0x1

            def foo(self):
                return "foo"

        assert Flag().foo() == "foo"

    def test_init_flag_type_allows_classmethods(self):
        class Flag(enums.Flag):
            A = 0x1

            @classmethod
            def foo(cls):
                assert cls is Flag
                return "foo"

        assert Flag.foo() == "foo"

    def test_init_flag_type_allows_staticmethods(self):
        class Flag(enums.Flag):
            A = 0x1

            @staticmethod
            def foo():
                return "foo"

        assert Flag.foo() == "foo"

    def test_init_flag_type_allows_descriptors(self):
        class Flag(enums.Flag):
            A = 0x1

            @property
            def foo(self):
                return "foo"

        assert isinstance(Flag.foo, property)

    def test_name_to_member_map(self):
        class Flag(enums.Flag):
            foo = 9
            bar = 18
            baz = 27

            @staticmethod
            def sm():
                pass

            @classmethod
            def cm(cls):
                pass

            def m(self):
                pass

            @property
            def p(self):
                pass

        assert Flag._name_to_member_map_["foo"].__class__ is Flag
        assert Flag._name_to_member_map_["foo"] is Flag.foo

        assert Flag._name_to_member_map_["bar"].__class__ is Flag
        assert Flag._name_to_member_map_["bar"] is Flag.bar

        assert Flag._name_to_member_map_["baz"].__class__ is Flag
        assert Flag._name_to_member_map_["baz"] is Flag.baz

        assert len(Flag._name_to_member_map_) == 3

    def test_value_to_member_map(self):
        class Flag(enums.Flag):
            foo = 9
            bar = 18
            baz = 27

            @staticmethod
            def sm():
                pass

            @classmethod
            def cm(cls):
                pass

            def m(self):
                pass

            @property
            def p(self):
                pass

        assert Flag._value_to_member_map_[9].__class__ is Flag
        assert Flag._value_to_member_map_[9] is Flag.foo

        assert Flag._value_to_member_map_[18].__class__ is Flag
        assert Flag._value_to_member_map_[18] is Flag.bar

        assert Flag._value_to_member_map_[27].__class__ is Flag
        assert Flag._value_to_member_map_[27] is Flag.baz

    def test_member_names(self):
        class Flag(enums.Flag):
            foo = 9
            bar = 18
            baz = 27

            @staticmethod
            def sm():
                pass

            @classmethod
            def cm(cls):
                pass

            def m(self):
                pass

            @property
            def p(self):
                pass

        assert Flag._member_names_ == ["foo", "bar", "baz"]

    def test_members(self):
        class Flag(enums.Flag):
            foo = 9
            bar = 18
            baz = 27

            @staticmethod
            def sm():
                pass

            @classmethod
            def cm(cls):
                pass

            def m(self):
                pass

            @property
            def p(self):
                pass

        assert Flag.__members__["foo"].__class__ is int
        assert Flag.__members__["foo"] == 9

        assert Flag.__members__["bar"].__class__ is int
        assert Flag.__members__["bar"] == 18

        assert Flag.__members__["baz"].__class__ is int
        assert Flag.__members__["baz"] == 27

        assert len(Flag.__members__) == 3

    def test_call_on_existing_value(self):
        class Flag(enums.Flag):
            foo = 9
            bar = 18
            baz = 27

        assert Flag(9) is Flag.foo
        assert Flag(Flag.foo) is Flag.foo

    def test_call_on_composite_value(self):
        class Flag(enums.Flag):
            foo = 1
            bar = 2
            baz = 4

        assert Flag(3) is Flag.foo | Flag.bar
        assert Flag(Flag.foo | Flag.bar) is Flag.foo | Flag.bar

    def test_call_on_named_composite_value(self):
        class Flag(enums.Flag):
            foo = 1
            bar = 2
            baz = 3

        assert Flag(3) is Flag.baz
        assert Flag(Flag.foo | Flag.bar) is Flag.baz

    def test_call_on_invalid_value(self):
        class Flag(enums.Flag):
            foo = 1
            bar = 2
            baz = 3

        assert Flag(4) == 4

    def test_cache(self):
        class Flag(enums.Flag):
            foo = 1
            bar = 2
            baz = 4

        assert Flag._temp_members_ == {}
        # Cache something. Remember the dict is evaluated before the equality
        # so this will populate the cache.
        assert Flag._temp_members_ == {3: Flag.foo | Flag.bar}
        assert Flag._temp_members_ == {3: Flag.foo | Flag.bar, 7: Flag.foo | Flag.bar | Flag.baz}

        # Shouldn't mutate for existing items.
        assert Flag._temp_members_ == {3: Flag.foo | Flag.bar, 7: Flag.foo | Flag.bar | Flag.baz}
        assert Flag._temp_members_ == {3: Flag.foo | Flag.bar, 7: Flag.foo | Flag.bar | Flag.baz}

    def test_cache_when_temp_values_over_MAX_CACHED_MEMBERS(self):
        class MockDict:
            def __getitem__(self, key):
                raise KeyError

            def __len__(self):
                return enums._MAX_CACHED_MEMBERS + 1

            def __setitem__(self, k, v):
                pass

            popitem = mock.Mock()

        class Flag(enums.Flag):
            foo = 1
            bar = 2
            baz = 3

        Flag._temp_members_ = MockDict()

        Flag(4)
        Flag._temp_members_.popitem.assert_called_once_with()

    def test_bitwise_name(self):
        class Flag(enums.Flag):
            foo = 1
            bar = 2
            baz = 4

        assert Flag.foo.name == "foo"

    def test_combined_bitwise_name(self):
        class Flag(enums.Flag):
            foo = 1
            bar = 2
            baz = 4

        assert (Flag.foo | Flag.bar).name == "foo|bar"

    def test_combined_known_bitwise_name(self):
        class Flag(enums.Flag):
            foo = 1
            bar = 2
            baz = 3

        assert (Flag.foo | Flag.bar).name == "baz"

    def test_combined_partially_known_name(self):
        class Flag(enums.Flag):
            doo = 1
            laa = 2
            dee = 3

        # This is fine because it is not an identity or exact value.
        assert (Flag.laa | 4 | Flag.doo).name == "doo|laa|0x4"

    def test_combined_partially_known_combined_bitwise_name(self):
        class Flag(enums.Flag):
            foo = 1
            bar = 2
            baz = 3

        # This is fine because it is not an identity or exact value.
        assert (Flag.baz | 4).name == "foo|bar|0x4"

    def test_unknown_name(self):
        class Flag(enums.Flag):
            foo = 1
            bar = 2
            baz = 3

        assert Flag(4).name == "UNKNOWN 0x4"

    def test_value(self):
        class Flag(enums.Flag):
            foo = 1
            bar = 2
            baz = 3

        assert Flag.foo.value.__class__ is int
        assert Flag.foo.value == 1

        assert Flag.bar.value.__class__ is int
        assert Flag.bar.value == 2

        assert Flag.baz.value.__class__ is int
        assert Flag.baz.value == 3

    def test_is_instance_of_declaring_type(self):
        class TestFlag(enums.Flag):
            FOO = 0x1
            BAR = 0x2
            BAZ = 0x3
            BORK = 0x4
            QUX = 0x8
            QUXX = QUX | BORK

        assert isinstance(TestFlag.BORK, TestFlag)
        assert isinstance(TestFlag.BORK, int)

        assert isinstance(TestFlag.QUXX, TestFlag)
        assert isinstance(TestFlag.QUXX, int)

    def test_and(self):
        class TestFlag(enums.Flag):
            FOO = 0x1
            BAR = 0x2
            BAZ = 0x3
            BORK = 0x4
            QUX = 0x8
            QUXX = QUX | BORK

        assert TestFlag.QUXX & TestFlag.QUX == TestFlag.QUX
        assert TestFlag.QUXX & TestFlag.QUX == 0x8
        assert TestFlag.QUXX & 0x8 == 0x8
        assert isinstance(TestFlag.QUXX & TestFlag.QUX, TestFlag)
        assert isinstance(TestFlag.QUXX & TestFlag.QUX, int)

    def test_rand(self):
        class TestFlag(enums.Flag):
            FOO = 0x1
            BAR = 0x2
            BAZ = 0x3
            BORK = 0x4
            QUX = 0x8
            QUXX = QUX | BORK

        assert 0x8 & TestFlag.QUXX == TestFlag.QUX
        assert 0x8 & TestFlag.QUXX == 0x8
        assert isinstance(0x8 & TestFlag.QUXX, TestFlag)
        assert isinstance(0x8 & TestFlag.QUXX, int)

    def test_all_positive_case(self):
        class TestFlag(enums.Flag):
            FOO = 0x1
            BAR = 0x2
            BAZ = 0x3
            BORK = 0x4
            QUX = 0x8

        val = TestFlag.BAZ | TestFlag.BORK

        assert val.all(TestFlag.FOO)

        assert val.all(TestFlag.FOO, TestFlag.BAR, TestFlag.BAZ, TestFlag.BORK)

    def test_all_negative_case(self):
        class TestFlag(enums.Flag):
            FOO = 0x1
            BAR = 0x2
            BAZ = 0x3
            BORK = 0x4
            QUX = 0x8
            QUXX = 0x10

        val = TestFlag.BAZ | TestFlag.BORK

        assert not val.all(TestFlag.QUX)
        assert not val.all(TestFlag.BAZ, TestFlag.QUX, TestFlag.QUXX)

    def test_any_positive_case(self):
        class TestFlag(enums.Flag):
            FOO = 0x1
            BAR = 0x2
            BAZ = 0x3
            BORK = 0x4
            QUX = 0x8

        val = TestFlag.BAZ | TestFlag.BORK

        assert val.any(TestFlag.FOO)
        assert val.any(TestFlag.BAR)
        assert val.any(TestFlag.BAZ)
        assert val.any(TestFlag.BORK)
        # All present
        assert val.any(TestFlag.FOO, TestFlag.BAR, TestFlag.BAZ, TestFlag.BORK)
        # One present, one not
        assert val.any(TestFlag.FOO, TestFlag.QUX)

    def test_any_negative_case(self):
        class TestFlag(enums.Flag):
            FOO = 0x1
            BAR = 0x2
            BAZ = 0x3
            BORK = 0x4
            QUX = 0x8
            QUXX = 0x10

        val = TestFlag.BAZ | TestFlag.BORK

        assert not val.any(TestFlag.QUX)

    def test_bool(self):
        class TestFlag(enums.Flag):
            BLEH = 0x0
            FOO = 0x1
            BAR = 0x2

        assert not TestFlag.BLEH
        assert TestFlag.FOO
        assert TestFlag.BAR

    def test_contains(self):
        class TestFlag(enums.Flag):
            BLEH = 0x1
            FOO = 0x2
            BAR = 0x4
            BAZ = 0x8

        f = TestFlag.FOO | TestFlag.BLEH | TestFlag.BAZ
        assert TestFlag.FOO in f
        assert TestFlag.BLEH in f
        assert TestFlag.BAZ in f
        assert TestFlag.BAR not in f

    def test_difference(self):
        class TestFlag(enums.Flag):
            A = 0x1
            B = 0x2
            C = 0x4
            D = 0x8
            E = 0x10
            F = 0x20
            G = 0x40
            H = 0x80

        a = TestFlag.A | TestFlag.B | TestFlag.D | TestFlag.F | TestFlag.G
        b = TestFlag.A | TestFlag.B | TestFlag.E
        c = 0x13
        expect_asubb = TestFlag.D | TestFlag.F | TestFlag.G
        expect_bsuba = TestFlag.E
        expect_asubc = 0x68

        assert a.difference(b) == expect_asubb
        assert b.difference(a) == expect_bsuba
        assert a.difference(c) == expect_asubc

        assert isinstance(a.difference(b), int)
        assert isinstance(a.difference(b), TestFlag)
        assert isinstance(a.difference(c), int)
        assert isinstance(a.difference(c), TestFlag)

    def test_index(self):
        class TestFlag(enums.Flag):
            OK = 0x5
            FOO = 0x312
            BAT = 0x3123

        assert operator.index(TestFlag.OK) == 0x5
        assert operator.index(TestFlag.FOO) == 0x312
        assert operator.index(TestFlag.BAT) == 0x3123

    def test_int(self):
        class TestFlag(enums.Flag):
            BLEH = 0x0
            FOO = 0x1
            BAR = 0x2

        assert int(TestFlag.BLEH) == 0x0
        assert int(TestFlag.FOO) == 0x1
        assert int(TestFlag.BAR) == 0x2

        assert type(int(TestFlag.BAR)) is int

    def test_intersection(self):
        class TestFlag(enums.Flag):
            FOO = 0x1
            BAR = 0x2
            BAZ = 0x3
            BORK = 0x4
            QUX = 0x8
            QUXX = QUX | BORK

        assert TestFlag.QUXX.intersection(TestFlag.QUX) == TestFlag.QUX
        assert TestFlag.QUXX.intersection(TestFlag.QUX) == 0x8
        assert TestFlag.QUXX.intersection(0x8) == 0x8
        assert isinstance(TestFlag.QUXX.intersection(TestFlag.QUX), TestFlag)
        assert isinstance(TestFlag.QUXX.intersection(TestFlag.QUX), int)

    def test_invert(self):
        class TestFlag(enums.Flag):
            FOO = 0x1
            BAR = 0x2
            BAZ = 0x4

        assert TestFlag.BAR.invert() == TestFlag.FOO | TestFlag.BAZ

    def test_invert_op(self):
        class TestFlag(enums.Flag):
            FOO = 0x1
            BAR = 0x2
            BAZ = 0x4

        assert ~TestFlag.BAR == TestFlag.FOO | TestFlag.BAZ

    def test_is_disjoint(self):
        class TestFlag(enums.Flag):
            FOO = 0x1
            BAR = 0x2
            BAZ = 0x4
            BORK = 0x8
            QUX = 0x10

        assert (TestFlag.FOO | TestFlag.BAR).is_disjoint(TestFlag.BAZ | TestFlag.BORK)
        assert not (TestFlag.FOO | TestFlag.BAR).is_disjoint(TestFlag.BAR | TestFlag.BORK)
        assert (TestFlag.FOO | TestFlag.BAR).is_disjoint(0xC)
        assert not (TestFlag.FOO | TestFlag.BAR).is_disjoint(0xA)

    def test_isdisjoint(self):
        class TestFlag(enums.Flag):
            FOO = 0x1
            BAR = 0x2
            BAZ = 0x4
            BORK = 0x8
            QUX = 0x10

        assert (TestFlag.FOO | TestFlag.BAR).isdisjoint(TestFlag.BAZ | TestFlag.BORK)
        assert not (TestFlag.FOO | TestFlag.BAR).isdisjoint(TestFlag.BAR | TestFlag.BORK)
        assert (TestFlag.FOO | TestFlag.BAR).isdisjoint(0xC)
        assert not (TestFlag.FOO | TestFlag.BAR).isdisjoint(0xA)

    def test_is_subset(self):
        class TestFlag(enums.Flag):
            BLEH = 0x1
            FOO = 0x2
            BAR = 0x4
            BAZ = 0x8
            BORK = 0x10

        f = TestFlag.FOO | TestFlag.BLEH | TestFlag.BAZ
        assert f.is_subset(TestFlag.FOO)
        assert f.is_subset(TestFlag.BLEH)
        assert f.is_subset(TestFlag.BAZ)
        assert not f.is_subset(TestFlag.BAR)
        assert f.is_subset(0x2)
        assert f.is_subset(0x1)
        assert f.is_subset(0x8)
        assert not f.is_subset(0x4)
        assert f.is_subset(TestFlag.FOO | TestFlag.BLEH)
        assert f.is_subset(0x3)
        assert f.is_subset(TestFlag.FOO | TestFlag.BLEH)
        assert not f.is_subset(TestFlag.BAR | TestFlag.BORK)
        assert not f.is_subset(0x14)

    def test_issubset(self):
        class TestFlag(enums.Flag):
            BLEH = 0x1
            FOO = 0x2
            BAR = 0x4
            BAZ = 0x8
            BORK = 0x10

        f = TestFlag.FOO | TestFlag.BLEH | TestFlag.BAZ
        assert f.issubset(TestFlag.FOO)
        assert f.issubset(TestFlag.BLEH)
        assert f.issubset(TestFlag.BAZ)
        assert not f.issubset(TestFlag.BAR)
        assert f.issubset(0x2)
        assert f.issubset(0x1)
        assert f.issubset(0x8)
        assert not f.issubset(0x4)
        assert f.issubset(TestFlag.FOO | TestFlag.BLEH)
        assert f.issubset(0x3)
        assert not f.issubset(TestFlag.BAR | TestFlag.BORK)
        assert not f.issubset(0x14)

    def test_is_superset(self):
        class TestFlag(enums.Flag):
            BLEH = 0x1
            FOO = 0x2
            BAR = 0x4
            BAZ = 0x8
            BORK = 0x10
            QUX = 0x10

        f = TestFlag.FOO | TestFlag.BLEH | TestFlag.BAZ

        assert f.is_superset(TestFlag.BLEH | TestFlag.FOO | TestFlag.BAR | TestFlag.BAZ | TestFlag.BORK)
        assert f.is_superset(0x1F)
        assert not f.is_superset(TestFlag.QUX)
        assert not f.is_superset(0x20)

    def test_issuperset(self):
        class TestFlag(enums.Flag):
            BLEH = 0x1
            FOO = 0x2
            BAR = 0x4
            BAZ = 0x8
            BORK = 0x10
            QUX = 0x10

        f = TestFlag.FOO | TestFlag.BLEH | TestFlag.BAZ

        assert f.issuperset(TestFlag.BLEH | TestFlag.FOO | TestFlag.BAR | TestFlag.BAZ | TestFlag.BORK)
        assert f.issuperset(0x1F)
        assert not f.issuperset(TestFlag.QUX)
        assert not f.issuperset(0x20)

    def test_iter(self):
        class TestFlag(enums.Flag):
            FOO = 0x1
            BAR = 0x2
            BAZ = 0x3
            BORK = 0x4
            QUX = 0x8

        val_iter = iter(TestFlag)
        assert next(val_iter) == TestFlag.FOO
        assert next(val_iter) == TestFlag.BAR
        assert next(val_iter) == TestFlag.BAZ
        assert next(val_iter) == TestFlag.BORK
        assert next(val_iter) == TestFlag.QUX
        with pytest.raises(StopIteration):
            next(val_iter)

    def test_flag_iter(self):
        class TestFlag(enums.Flag):
            FOO = 0x1
            BAR = 0x2
            BAZ = 0x3
            BORK = 0x4
            QUX = 0x8

        val = TestFlag.BAZ | TestFlag.BORK
        val_iter = iter(val)
        assert next(val_iter) == TestFlag.BAR
        assert next(val_iter) == TestFlag.BORK
        assert next(val_iter) == TestFlag.FOO
        with pytest.raises(StopIteration):
            next(val_iter)

    def test_len(self):
        class TestFlag(enums.Flag):
            FOO = 0x1
            BAR = 0x2
            BAZ = 0x3
            BORK = 0x4
            QUX = 0x8

        val0 = TestFlag(0)
        val1 = TestFlag.FOO
        val2 = TestFlag.FOO | TestFlag.BORK
        val3 = TestFlag.FOO | TestFlag.BAR | TestFlag.BORK
        val3_comb = TestFlag.BAZ | TestFlag.BORK

        assert len(val0) == 0
        assert len(val1) == 1
        assert len(val2) == 2
        assert len(val3) == 3
        assert len(val3_comb) == 3

    def test_or(self):
        class TestFlag(enums.Flag):
            FOO = 0x1
            BAR = 0x2
            BAZ = 0x4
            BORK = 0x8
            QUX = 0x10

        assert isinstance(TestFlag.FOO | TestFlag.BAR, int)
        assert isinstance(TestFlag.FOO | TestFlag.BAR, TestFlag)
        assert isinstance(TestFlag.FOO | 0x2, int)
        assert isinstance(TestFlag.FOO | 0x2, TestFlag)

        assert TestFlag.FOO | TestFlag.BAR == 0x3
        assert TestFlag.FOO | TestFlag.BAR == TestFlag(0x3)
        assert TestFlag.FOO | 0x2 == 0x3
        assert TestFlag.FOO | 0x2 == TestFlag(0x3)

    def test_ror(self):
        class TestFlag(enums.Flag):
            FOO = 0x1
            BAR = 0x2
            BAZ = 0x4
            BORK = 0x8
            QUX = 0x10

        assert isinstance(0x2 | TestFlag.FOO, int)
        assert isinstance(0x2 | TestFlag.FOO, TestFlag)

        assert 0x2 | TestFlag.FOO == 0x3
        assert 0x2 | TestFlag.FOO == TestFlag(0x3)

    def test_none_positive_case(self):
        class TestFlag(enums.Flag):
            FOO = 0x1
            BAR = 0x2
            BAZ = 0x3
            BORK = 0x4
            QUX = 0x8
            QUXX = 0x10

        val = TestFlag.BAZ | TestFlag.BORK

        assert val.none(TestFlag.QUX)

    def test_none_negative_case(self):
        class TestFlag(enums.Flag):
            FOO = 0x1
            BAR = 0x2
            BAZ = 0x3
            BORK = 0x4
            QUX = 0x8

        val = TestFlag.BAZ | TestFlag.BORK

        assert not val.none(TestFlag.FOO)
        assert not val.none(TestFlag.BAR)
        assert not val.none(TestFlag.BAZ)
        assert not val.none(TestFlag.BORK)
        # All present
        assert not val.none(TestFlag.FOO, TestFlag.BAR, TestFlag.BAZ, TestFlag.BORK)
        # One present, one not
        assert not val.none(TestFlag.FOO, TestFlag.QUX)

    def test_split(self):
        class TestFlag(enums.Flag):
            FOO = 0x1
            BAR = 0x2
            BAZ = 0x3
            BORK = 0x4
            QUX = 0x8

        val = TestFlag.BAZ | TestFlag.BORK

        # Baz is a combined field technically, so we don't expect it to be output here
        assert val.split() == [TestFlag.BAR, TestFlag.BORK, TestFlag.FOO]

    def test_str_operator(self):
        class TestFlag(enums.Flag):
            FOO = 0x1
            BAR = 0x2
            BAZ = 0x3
            BORK = 0x4
            QUX = 0x8

        val = TestFlag.BAZ | TestFlag.BORK

        assert str(val) == "FOO|BAR|BORK"

    def test_symmetric_difference(self):
        class TestFlag(enums.Flag):
            FOO = 0x1
            BAR = 0x2
            BAZ = 0x4
            BORK = 0x8
            QUX = 0x10

        a = TestFlag.FOO | TestFlag.BAR | TestFlag.BAZ
        b = TestFlag.BAZ | TestFlag.BORK | TestFlag.QUX

        assert isinstance(a.symmetric_difference(b), int)
        assert isinstance(a.symmetric_difference(b), TestFlag)
        assert isinstance(a.symmetric_difference(0x1C), int)
        assert isinstance(a.symmetric_difference(0x1C), TestFlag)

        assert a.symmetric_difference(b) == b.symmetric_difference(a)
        assert a.symmetric_difference(a) == 0
        assert b.symmetric_difference(b) == 0

        assert a.symmetric_difference(b) == TestFlag.FOO | TestFlag.BAR | TestFlag.BORK | TestFlag.QUX
        assert a.symmetric_difference(b) == 0x1B

    def test_sub(self):
        class TestFlag(enums.Flag):
            A = 0x1
            B = 0x2
            C = 0x4
            D = 0x8
            E = 0x10
            F = 0x20
            G = 0x40
            H = 0x80

        a = TestFlag.A | TestFlag.B | TestFlag.D | TestFlag.F | TestFlag.G
        b = TestFlag.A | TestFlag.B | TestFlag.E
        c = 0x13
        expect_asubb = TestFlag.D | TestFlag.F | TestFlag.G
        expect_bsuba = TestFlag.E
        expect_asubc = 0x68

        assert a - b == expect_asubb
        assert b - a == expect_bsuba
        assert a - c == expect_asubc

        assert isinstance(a - b, int)
        assert isinstance(a - b, TestFlag)
        assert isinstance(a - c, int)
        assert isinstance(a - c, TestFlag)

    def test_rsub(self):
        class TestFlag(enums.Flag):
            A = 0x1
            B = 0x2
            C = 0x4
            D = 0x8
            E = 0x10
            F = 0x20
            G = 0x40
            H = 0x80

        a = TestFlag.A | TestFlag.B | TestFlag.D | TestFlag.F | TestFlag.G
        c = 0x13
        expect_csuba = 0x10

        assert c - a == expect_csuba

        assert isinstance(c - a, int)
        assert isinstance(c - a, TestFlag)

    def test_union(self):
        class TestFlag(enums.Flag):
            FOO = 0x1
            BAR = 0x2
            BAZ = 0x4
            BORK = 0x8
            QUX = 0x10

        assert isinstance(TestFlag.FOO.union(TestFlag.BAR), int)
        assert isinstance(TestFlag.FOO.union(TestFlag.BAR), TestFlag)
        assert isinstance(TestFlag.FOO.union(TestFlag.BAR).union(TestFlag.BAZ), int)
        assert isinstance(TestFlag.FOO.union(TestFlag.BAR).union(TestFlag.BAZ), TestFlag)
        assert isinstance(TestFlag.FOO.union(0x2).union(TestFlag.BAZ), int)
        assert isinstance(TestFlag.FOO.union(0x2).union(TestFlag.BAZ), TestFlag)
        assert isinstance(TestFlag.FOO.union(0x2), int)
        assert isinstance(TestFlag.FOO.union(0x2), TestFlag)

        assert TestFlag.FOO.union(TestFlag.BAR) == 0x3
        assert TestFlag.FOO.union(TestFlag.BAR) == TestFlag(0x3)
        assert TestFlag.FOO.union(0x2) == 0x3
        assert TestFlag.FOO.union(0x2) == TestFlag(0x3)

    def test_xor(self):
        class TestFlag(enums.Flag):
            FOO = 0x1
            BAR = 0x2
            BAZ = 0x4
            BORK = 0x8
            QUX = 0x10

        a = TestFlag.FOO | TestFlag.BAR | TestFlag.BAZ
        b = TestFlag.BAZ | TestFlag.BORK | TestFlag.QUX

        assert isinstance(a ^ b, int)
        assert isinstance(a ^ b, TestFlag)
        assert isinstance(a ^ 0x1C, int)
        assert isinstance(a ^ 0x1C, TestFlag)

        assert a ^ b == b ^ a
        assert a ^ a == 0
        assert b ^ b == 0

        assert a ^ b == TestFlag.FOO | TestFlag.BAR | TestFlag.BORK | TestFlag.QUX
        assert a ^ b == 0x1B
        assert a ^ 0x1C == TestFlag.FOO | TestFlag.BAR | TestFlag.BORK | TestFlag.QUX
        assert a ^ 0x1C == 0x1B

    def test_rxor(self):
        class TestFlag(enums.Flag):
            FOO = 0x1
            BAR = 0x2
            BAZ = 0x4
            BORK = 0x8
            QUX = 0x10

        a = TestFlag.FOO | TestFlag.BAR | TestFlag.BAZ

        assert isinstance(0x1C ^ a, int)
        assert isinstance(0x1C ^ a, TestFlag)
        assert 0x1C ^ a == TestFlag.FOO | TestFlag.BAR | TestFlag.BORK | TestFlag.QUX
        assert 0x1C ^ a == 0x1B

    def test_getitem(self):
        class TestFlag(enums.Flag):
            FOO = 0x1
            BAR = 0x2
            BAZ = 0x4
            BORK = 0x8
            QUX = 0x10

        returned = TestFlag["FOO"]
        assert returned == TestFlag.FOO
        assert type(returned) is TestFlag

    def test_repr(self):
        class TestFlag(enums.Flag):
            FOO = 0x1
            BAR = 0x2
            BAZ = 0x4
            BORK = 0x8
            QUX = 0x10

        assert repr(TestFlag) == "<enum TestFlag>"
        assert repr(TestFlag.FOO) == "<TestFlag.FOO: 1>"

    def test_allows_overriding_methods(self):
        class TestFlag(enums.Flag):
            FOO = 0x1
            BAR = 0x2
            BAZ = 0x4
            BORK = 0x8
            QUX = 0x10

            def __int__(self):
                return 855555

        assert int(TestFlag.FOO | TestFlag.BAR) == 855555


def test_deprecated():
    with mock.patch.object(deprecation, "check_if_past_removal"):

        class Enum(int, enums.Enum):
            OK_VALUE = 1
            DEPRECATED = enums.deprecated(OK_VALUE, removal_version="4.0.0")

        with mock.patch.object(warnings, "warn") as warn:
            assert Enum.DEPRECATED == Enum.OK_VALUE
            warn.assert_called_once()
            warn.reset_mock()

            assert Enum["DEPRECATED"] == Enum.OK_VALUE
            warn.assert_called_once()
            warn.reset_mock()

            assert Enum.DEPRECATED.value == Enum.OK_VALUE.value
            warn.assert_called_once()
            warn.reset_mock()

            # Ensure we didn't break any other attributes
            assert Enum(1) == Enum.OK_VALUE
            warn.assert_not_called()

            Enum.OK_VALUE
            warn.assert_not_called()



# Copyright (c) 2020 Nekokatt
# Copyright (c) 2021-present davfsa
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import annotations

import mock
import pytest

from hikari import audit_logs
from hikari import channels
from hikari import snowflakes


@pytest.mark.asyncio
class TestMessagePinEntryInfo:
    async def test_fetch_channel(self):
        app = mock.AsyncMock()
        app.rest.fetch_channel.return_value = mock.Mock(spec_set=channels.GuildTextChannel)
        model = audit_logs.MessagePinEntryInfo(app=app, channel_id=123, message_id=456)

        assert await model.fetch_channel() is model.app.rest.fetch_channel.return_value

        model.app.rest.fetch_channel.assert_awaited_once_with(123)

    async def test_fetch_message(self):
        model = audit_logs.MessagePinEntryInfo(app=mock.AsyncMock(), channel_id=123, message_id=456)

        assert await model.fetch_message() is model.app.rest.fetch_message.return_value

        model.app.rest.fetch_message.assert_awaited_once_with(123, 456)


@pytest.mark.asyncio
class TestMessageDeleteEntryInfo:
    async def test_fetch_channel(self):
        app = mock.AsyncMock()
        app.rest.fetch_channel.return_value = mock.Mock(spec_set=channels.GuildTextChannel)
        model = audit_logs.MessageDeleteEntryInfo(app=app, count=1, channel_id=123)

        assert await model.fetch_channel() is model.app.rest.fetch_channel.return_value

        model.app.rest.fetch_channel.assert_awaited_once_with(123)


@pytest.mark.asyncio
class TestMemberMoveEntryInfo:
    async def test_fetch_channel(self):
        app = mock.AsyncMock()
        app.rest.fetch_channel.return_value = mock.Mock(spec_set=channels.GuildVoiceChannel)
        model = audit_logs.MemberMoveEntryInfo(app=app, count=1, channel_id=123)

        assert await model.fetch_channel() is model.app.rest.fetch_channel.return_value

        model.app.rest.fetch_channel.assert_awaited_once_with(123)


class TestAuditLogEntry:
    @pytest.mark.asyncio
    async def test_fetch_user_when_no_user(self):
        model = audit_logs.AuditLogEntry(
            app=mock.AsyncMock(),
            id=123,
            target_id=None,
            changes=[],
            user_id=None,
            action_type=0,
            options=None,
            reason=None,
            guild_id=snowflakes.Snowflake(34123123),
        )

        assert await model.fetch_user() is None

        model.app.rest.fetch_user.assert_not_called()

    @pytest.mark.asyncio
    async def test_fetch_user_when_user(self):
        model = audit_logs.AuditLogEntry(
            app=mock.AsyncMock(),
            id=123,
            target_id=None,
            changes=[],
            user_id=456,
            action_type=0,
            options=None,
            reason=None,
            guild_id=snowflakes.Snowflake(123321123),
        )

        assert await model.fetch_user() is model.app.rest.fetch_user.return_value

        model.app.rest.fetch_user.assert_awaited_once_with(456)


class TestAuditLog:
    def test_iter(self):
        entry_1 = object()
        entry_2 = object()
        entry_3 = object()
        audit_log = audit_logs.AuditLog(
            auto_mod_rules={},
            entries={
                snowflakes.Snowflake(432123): entry_1,
                snowflakes.Snowflake(432654): entry_2,
                snowflakes.Snowflake(432888): entry_3,
            },
            integrations={},
            users={},
            threads={},
            webhooks={},
        )
        assert list(audit_log) == [entry_1, entry_2, entry_3]

    def test_get_item_with_index(self):
        entry = object()
        entry_2 = object()
        audit_log = audit_logs.AuditLog(
            auto_mod_rules={},
            entries={
                snowflakes.Snowflake(432123): object(),
                snowflakes.Snowflake(432654): entry,
                snowflakes.Snowflake(432888): object(),
                snowflakes.Snowflake(677777): object(),
                snowflakes.Snowflake(999999): entry_2,
            },
            integrations={},
            threads={},
            users={},
            webhooks={},
        )
        assert audit_log[1] is entry
        assert audit_log[4] is entry_2

    def test_get_item_with_slice(self):
        entry_1 = object()
        entry_2 = object()
        audit_log = audit_logs.AuditLog(
            auto_mod_rules={},
            entries={
                snowflakes.Snowflake(432123): object(),
                snowflakes.Snowflake(432654): entry_1,
                snowflakes.Snowflake(432888): object(),
                snowflakes.Snowflake(666666): entry_2,
                snowflakes.Snowflake(783452): object(),
            },
            integrations={},
            threads={},
            users={},
            webhooks={},
        )
        assert audit_log[1:5:2] == (entry_1, entry_2)

    def test_len(self):
        audit_log = audit_logs.AuditLog(
            auto_mod_rules={},
            entries={
                snowflakes.Snowflake(432123): object(),
                snowflakes.Snowflake(432654): object(),
                snowflakes.Snowflake(432888): object(),
                snowflakes.Snowflake(783452): object(),
            },
            integrations={},
            threads={},
            users={},
            webhooks={},
        )
        assert len(audit_log) == 4



# Copyright (c) 2020 Nekokatt
# Copyright (c) 2021-present davfsa
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Test twemoji mapping.

A CI-used script that tests all the Twemoji URLs generated
by Discord emojis are actually legitimate URLs, since Discord
does not map these on a 1-to-1 basis.
"""

from __future__ import annotations

import json
import pathlib
import shutil
import subprocess
import sys
import time
import urllib.request

sys.path.append("..")


from hikari import emojis

TWEMOJI_REPO_BASE_URL = "https://github.com/discord/twemoji.git"
DISCORD_EMOJI_MAPPING_URL = "https://emzi0767.mzgit.io/discord-emoji/discordEmojiMap-canary.min.json"

try:
    tempdir = pathlib.Path(sys.argv[1])
except IndexError:
    print("Argument 1 must be the path to the temporary directory")
    sys.exit(2)

git = shutil.which("git")

if not git:
    print("Git not detected, please ensure it is installed properly before continuing")
    sys.exit(3)

start = time.perf_counter()

print("Fetching emoji mapping")
with urllib.request.urlopen(DISCORD_EMOJI_MAPPING_URL) as request:  # noqa: S310
    mapping = json.loads(request.read())["emojiDefinitions"]

has_items = next(tempdir.iterdir(), False)
if has_items:
    print("Updating twemoji collection")
    subprocess.check_call(f"{git} pull", shell=True, cwd=tempdir)
else:
    print("Cloning twemoji collection")
    subprocess.check_call(f"{git} clone {TWEMOJI_REPO_BASE_URL} {tempdir} --depth=1", shell=True)

assets_path = tempdir / "assets" / "72x72"

invalid_emojis: list[str] = []
total = len(mapping)
for i, emoji in enumerate(mapping, start=1):
    name = emoji["primaryName"]
    emoji_surrogates = emoji["surrogates"]
    hikari_emoji = emojis.UnicodeEmoji.parse(emoji_surrogates)
    line_repr = f"{i}/{total} {name} {' '.join(map(hex, map(ord, emoji_surrogates)))} {hikari_emoji.url}"

    if (assets_path / hikari_emoji.filename).exists():
        print(f"[  OK  ] {line_repr}")
    else:
        invalid_emojis.append(line_repr)
        print(f"[ FAIL ] {line_repr}")

print()
print("Results")
print("-------")
print(f"Valid emojis: {total - len(invalid_emojis)}")
print(f"Invalid emojis: {len(invalid_emojis)}")

for line_repr in invalid_emojis:
    print(f"  {line_repr}")

print(f"Took {time.perf_counter() - start} seconds")

if invalid_emojis or total == 0:
    sys.exit(1)



# Copyright (c) 2020 Nekokatt
# Copyright (c) 2021-present davfsa
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import annotations

import asyncio

import mock
import pytest

from hikari import errors
from hikari import snowflakes
from hikari.events import voice_events
from hikari.impl import voice
from tests.hikari import hikari_test_helpers


class TestVoiceComponentImpl:
    @pytest.fixture
    def mock_app(self):
        return mock.Mock()

    @pytest.fixture
    def voice_client(self, mock_app):
        client = hikari_test_helpers.mock_class_namespace(voice.VoiceComponentImpl, slots_=False)(mock_app)
        client._is_alive = True
        return client

    def test_is_alive_property(self, voice_client):
        voice_client.is_alive is voice_client._is_alive

    def test__check_if_alive_when_alive(self, voice_client):
        voice_client._is_alive = True
        voice_client._check_if_alive()

    def test__check_if_alive_when_not_alive(self, voice_client):
        voice_client._is_alive = False

        with pytest.raises(errors.ComponentStateConflictError):
            voice_client._check_if_alive()

    def test__check_if_alive_when_closing(self, voice_client):
        voice_client._is_alive = True
        voice_client._is_closing = True

        with pytest.raises(errors.ComponentStateConflictError):
            voice_client._check_if_alive()

    @pytest.mark.asyncio
    async def test_disconnect(self, voice_client):
        mock_connection = mock.AsyncMock()
        mock_connection_2 = mock.AsyncMock()
        voice_client._connections = {
            snowflakes.Snowflake(123): mock_connection,
            snowflakes.Snowflake(5324): mock_connection_2,
        }
        voice_client._check_if_alive = mock.Mock()

        await voice_client.disconnect(123)

        voice_client._check_if_alive.assert_called_once_with()
        mock_connection.disconnect.assert_awaited_once_with()
        mock_connection_2.disconnect.assert_not_called()

    @pytest.mark.asyncio
    async def test_disconnect_when_guild_id_not_in_connections(self, voice_client):
        mock_connection = mock.AsyncMock()
        mock_connection_2 = mock.AsyncMock()
        voice_client._connections = {123: mock_connection, 5324: mock_connection_2}

        with pytest.raises(errors.VoiceError):
            await voice_client.disconnect(1234567890)

        mock_connection.disconnect.assert_not_called()
        mock_connection_2.disconnect.assert_not_called()

    @pytest.mark.asyncio
    async def test__disconnect_all(self, voice_client):
        mock_connection = mock.AsyncMock()
        mock_connection_2 = mock.AsyncMock()
        voice_client._connections = {123: mock_connection, 5324: mock_connection_2}

        await voice_client._disconnect_all()

        mock_connection.disconnect.assert_awaited_once_with()
        mock_connection_2.disconnect.assert_awaited_once_with()

    @pytest.mark.asyncio
    async def test_disconnect_all(self, voice_client):
        voice_client._disconnect_all = mock.AsyncMock()
        voice_client._check_if_alive = mock.Mock()

        await voice_client.disconnect_all()

        voice_client._check_if_alive.assert_called_once_with()
        voice_client._disconnect_all.assert_awaited_once_with()

    @pytest.mark.asyncio
    @pytest.mark.parametrize("voice_listener", [True, False])
    async def test_close(self, voice_client, mock_app, voice_listener):
        voice_client._disconnect_all = mock.AsyncMock()
        voice_client._connections = {123: None}
        voice_client._check_if_alive = mock.Mock()
        voice_client._voice_listener = voice_listener

        await voice_client.close()

        if voice_listener:
            mock_app.event_manager.unsubscribe.assert_called_once_with(
                voice_events.VoiceEvent, voice_client._on_voice_event
            )
        else:
            mock_app.event_manager.unsubscribe.assert_not_called()

        voice_client._check_if_alive.assert_called_once_with()
        voice_client._disconnect_all.assert_awaited_once_with()
        assert voice_client._is_alive is False
        assert voice_client._is_closing is False

    @pytest.mark.asyncio
    @pytest.mark.parametrize("voice_listener", [True, False])
    async def test_close_when_no_connections(self, voice_client, mock_app, voice_listener):
        voice_client._disconnect_all = mock.AsyncMock()
        voice_client._connections = {}
        voice_client._check_if_alive = mock.Mock()
        voice_client._voice_listener = voice_listener

        await voice_client.close()

        if voice_listener:
            mock_app.event_manager.unsubscribe.assert_called_once_with(
                voice_events.VoiceEvent, voice_client._on_voice_event
            )
        else:
            mock_app.event_manager.unsubscribe.assert_not_called()

        voice_client._check_if_alive.assert_called_once_with()
        voice_client._disconnect_all.assert_not_called()
        assert voice_client._is_alive is False
        assert voice_client._is_closing is False

    def test_start(self, voice_client, mock_app):
        voice_client._is_alive = False

        voice_client.start()

        assert voice_client._is_alive is True

    @pytest.mark.asyncio
    async def test_start_when_already_alive(self, voice_client, mock_app):
        voice_client._is_alive = True

        with pytest.raises(errors.ComponentStateConflictError):
            await voice_client.start()

    @pytest.mark.asyncio
    @pytest.mark.parametrize("voice_listener", [True, False])
    async def test_connect_to(self, voice_client, mock_app, voice_listener):
        voice_client._init_state_update_predicate = mock.Mock()
        voice_client._init_server_update_predicate = mock.Mock()
        mock_other_connection = object()
        voice_client._connections = {555: mock_other_connection}
        mock_shard = mock.AsyncMock(is_alive=True)
        mock_app.event_manager.wait_for = mock.AsyncMock()
        mock_app.shard_count = 42
        mock_app.shards = {0: mock_shard}
        mock_connection_type = mock.AsyncMock()
        voice_client._check_if_alive = mock.Mock()
        voice_client._voice_listener = voice_listener

        result = await voice_client.connect_to(123, 4532, mock_connection_type, deaf=False, mute=True, timeout=None)

        voice_client._check_if_alive.assert_called_once_with()
        mock_app.event_manager.wait_for.assert_has_awaits(
            [
                mock.call(
                    voice_events.VoiceStateUpdateEvent,
                    timeout=None,
                    predicate=voice_client._init_state_update_predicate.return_value,
                ),
                mock.call(
                    voice_events.VoiceServerUpdateEvent,
                    timeout=None,
                    predicate=voice_client._init_server_update_predicate.return_value,
                ),
            ]
        )
        mock_app.rest.fetch_my_user.assert_not_called()
        mock_app.cache.get_me.assert_called_once_with()
        voice_client._init_state_update_predicate.assert_called_once_with(123, mock_app.cache.get_me.return_value.id)
        voice_client._init_server_update_predicate.assert_called_once_with(123)
        if voice_listener:
            mock_app.event_manager.subscribe.assert_not_called()
        else:
            mock_app.event_manager.subscribe.assert_called_once_with(
                voice_events.VoiceEvent, voice_client._on_voice_event
            )

        assert voice_client._voice_listener is True
        mock_shard.update_voice_state.assert_awaited_once_with(123, 4532, self_deaf=False, self_mute=True)
        assert voice_client._connections == {
            123: mock_connection_type.initialize.return_value,
            555: mock_other_connection,
        }
        assert result is mock_connection_type.initialize.return_value

    @pytest.mark.asyncio
    async def test_connect_to_fails_when_wait_for_timeout(self, voice_client, mock_app):
        mock_shard = mock.AsyncMock(is_alive=True)
        mock_wait_for = mock.AsyncMock()
        mock_wait_for.side_effect = asyncio.TimeoutError
        mock_app.event_manager.wait_for = mock_wait_for
        mock_app.shard_count = 42
        mock_app.shards = {0: mock_shard}
        mock_connection_type = mock.AsyncMock()

        with pytest.raises(errors.VoiceError, match="Could not connect to voice channel 4532 in guild 123."):
            await voice_client.connect_to(123, 4532, mock_connection_type)

    @pytest.mark.asyncio
    async def test_connect_to_falls_back_to_rest_to_get_own_user(self, voice_client, mock_app):
        voice_client._init_state_update_predicate = mock.Mock()
        voice_client._init_server_update_predicate = mock.Mock()
        mock_shard = mock.AsyncMock(is_alive=True)
        mock_app.event_manager.wait_for = mock.AsyncMock()
        mock_app.shard_count = 42
        mock_app.shards = {0: mock_shard}
        mock_app.cache.get_me.return_value = None
        mock_app.rest = mock.AsyncMock()
        mock_connection_type = mock.AsyncMock()

        await voice_client.connect_to(123, 4532, mock_connection_type, deaf=False, mute=True, timeout=None)

        mock_app.event_manager.wait_for.assert_has_awaits(
            [
                mock.call(
                    voice_events.VoiceStateUpdateEvent,
                    timeout=None,
                    predicate=voice_client._init_state_update_predicate.return_value,
                ),
                mock.call(
                    voice_events.VoiceServerUpdateEvent,
                    timeout=None,
                    predicate=voice_client._init_server_update_predicate.return_value,
                ),
            ]
        )
        mock_app.cache.get_me.assert_called_once_with()
        mock_app.rest.fetch_my_user.assert_awaited_once_with()
        voice_client._init_state_update_predicate.assert_called_once_with(
            123, mock_app.rest.fetch_my_user.return_value.id
        )

    @pytest.mark.asyncio
    async def test_connect_to_when_connection_already_present(self, voice_client, mock_app):
        voice_client._connections = {snowflakes.Snowflake(123): object()}

        with pytest.raises(
            errors.VoiceError,
            match="Already in a voice channel for that guild. Disconnect before attempting to connect again",
        ):
            await voice_client.connect_to(123, 4532, object())

    @pytest.mark.asyncio
    async def test_connect_to_for_unknown_shard(self, voice_client, mock_app):
        mock_app.shard_count = 42
        mock_app.shards = {}

        with pytest.raises(
            errors.VoiceError, match="Cannot connect to shard 0 as it is not present in this application"
        ):
            await voice_client.connect_to(123, 4532, object())

    @pytest.mark.asyncio
    async def test_connect_to_handles_failed_connection_initialise(self, voice_client, mock_app):
        voice_client._init_state_update_predicate = mock.Mock()
        voice_client._init_server_update_predicate = mock.Mock()
        mock_shard = mock.Mock(is_alive=True)
        update_voice_state_call_1 = mock.AsyncMock()
        update_voice_state_call_2 = mock.Mock()
        mock_shard.update_voice_state = mock.Mock(
            side_effect=[update_voice_state_call_1(), update_voice_state_call_2()]
        )
        mock_app.event_manager.wait_for = mock.AsyncMock()
        mock_app.shard_count = 42
        mock_app.shards = {0: mock_shard}

        class StubError(Exception): ...

        mock_connection_type = mock.AsyncMock()
        mock_connection_type.initialize.side_effect = StubError

        with mock.patch.object(
            asyncio, "wait_for", new=mock.AsyncMock(side_effect=asyncio.TimeoutError)
        ) as asyncio_wait_for:
            with pytest.raises(StubError):
                await voice_client.connect_to(123, 4532, mock_connection_type, deaf=False, mute=True, timeout=None)

        mock_app.event_manager.wait_for.assert_has_awaits(
            [
                mock.call(
                    voice_events.VoiceStateUpdateEvent,
                    timeout=None,
                    predicate=voice_client._init_state_update_predicate.return_value,
                ),
                mock.call(
                    voice_events.VoiceServerUpdateEvent,
                    timeout=None,
                    predicate=voice_client._init_server_update_predicate.return_value,
                ),
            ]
        )
        mock_app.cache.get_me.assert_called_once_with()
        voice_client._init_state_update_predicate.assert_called_once_with(123, mock_app.cache.get_me.return_value.id)
        voice_client._init_server_update_predicate.assert_called_once_with(123)
        mock_shard.update_voice_state.assert_has_calls(
            [mock.call(123, 4532, self_deaf=False, self_mute=True), mock.call(123, None)]
        )
        update_voice_state_call_1.assert_awaited_once()
        asyncio_wait_for.assert_awaited_once_with(update_voice_state_call_2.return_value, timeout=5.0)

    @pytest.mark.asyncio
    @pytest.mark.parametrize("more_connections", [True, False])
    async def test__on_connection_close(self, voice_client, mock_app, more_connections):
        mock_shard = mock.AsyncMock()
        mock_app.shards = {69: mock_shard}
        voice_client._connections = {65234123: object()}
        expected_connections = {}
        if more_connections:
            mock_connection = object()
            voice_client._connections[123] = mock_connection
            expected_connections[123] = mock_connection

        await voice_client._on_connection_close(mock.Mock(guild_id=65234123, shard_id=69))

        if more_connections:
            mock_app.event_manager.unsubscribe.assert_not_called()
        else:
            mock_app.event_manager.unsubscribe.assert_called_once_with(
                voice_events.VoiceEvent, voice_client._on_voice_event
            )

        mock_shard.update_voice_state.assert_awaited_once_with(guild=65234123, channel=None)
        assert voice_client._connections == expected_connections

    def test__init_state_update_predicate_matches(self, voice_client):
        predicate = voice_client._init_state_update_predicate(42069, 696969)
        mock_voice_state = mock.Mock(state=mock.Mock(guild_id=42069, user_id=696969))

        assert predicate(mock_voice_state) is True

    def test__init_state_update_predicate_ignores(self, voice_client):
        predicate = voice_client._init_state_update_predicate(999, 420)
        mock_voice_state = mock.Mock(state=mock.Mock(guild_id=6969, user_id=3333))

        assert predicate(mock_voice_state) is False

    def test__init_server_update_predicate_matches(self, voice_client):
        predicate = voice_client._init_server_update_predicate(696969)
        mock_voice_state = mock.Mock(guild_id=696969)

        assert predicate(mock_voice_state) is True

    def test__init_server_update_predicate_ignores(self, voice_client):
        predicate = voice_client._init_server_update_predicate(321231)
        mock_voice_state = mock.Mock(guild_id=123123123)

        assert predicate(mock_voice_state) is False

    @pytest.mark.asyncio
    async def test__on_connection_close_ignores_unknown_voice_state(self, voice_client):
        connections = {123132: object(), 65234234: object()}
        voice_client._connections = connections.copy()

        await voice_client._on_connection_close(mock.Mock(guild_id=-1))

        assert voice_client._connections == connections

    @pytest.mark.asyncio
    async def test__on_voice_event(self, voice_client):
        mock_connection = mock.AsyncMock()
        voice_client._connections = {6633: mock_connection}
        mock_event = mock.Mock(guild_id=6633)

        await voice_client._on_voice_event(mock_event)

        mock_connection.notify.assert_awaited_once_with(mock_event)

    @pytest.mark.asyncio
    async def test__on_voice_event_for_untracked_guild(self, voice_client):
        mock_event = mock.Mock(guild_id=44444)

        await voice_client._on_voice_event(mock_event)



# Copyright (c) 2020 Nekokatt
# Copyright (c) 2021-present davfsa
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import annotations

import mock
import pytest

from hikari import snowflakes
from hikari import stickers
from hikari import urls
from hikari.internal import routes


class TestStickerPack:
    @pytest.fixture
    def model(self):
        return stickers.StickerPack(
            id=123,
            name="testing",
            description="testing description",
            cover_sticker_id=snowflakes.Snowflake(6541234),
            stickers=[],
            sku_id=123,
            banner_asset_id=snowflakes.Snowflake(541231),
        )

    def test_make_banner_url_format_set_to_deprecated_ext_argument_if_provided(self, model):
        with mock.patch.object(
            routes, "CDN_STICKER_PACK_BANNER", new=mock.Mock(compile_to_file=mock.Mock(return_value="file"))
        ) as route:
            assert model.make_banner_url(ext="JPEG") == "file"

        route.compile_to_file.assert_called_once_with(
            urls.CDN_URL, hash=541231, size=4096, file_format="JPEG", lossless=True
        )

    def test_banner_url(self, model):
        banner = object()

        with mock.patch.object(stickers.StickerPack, "make_banner_url", return_value=banner):
            assert model.banner_url is banner

    def test_make_banner_url(self, model):
        with mock.patch.object(
            routes, "CDN_STICKER_PACK_BANNER", new=mock.Mock(compile_to_file=mock.Mock(return_value="file"))
        ) as route:
            assert model.make_banner_url(file_format="URL", size=512) == "file"

        route.compile_to_file.assert_called_once_with(
            urls.CDN_URL, hash=541231, size=512, file_format="URL", lossless=True
        )

    def test_make_banner_url_when_no_banner_asset(self, model):
        model.banner_asset_id = None

        assert model.make_banner_url(file_format="URL", size=512) is None


class TestPartialSticker:
    @pytest.fixture
    def model(self):
        return stickers.PartialSticker(id=123, name="testing", format_type="some")

    def test_make_url_uses_CDN_when_LOTTIE(self, model):
        model.format_type = stickers.StickerFormatType.LOTTIE

        with mock.patch.object(
            routes, "CDN_STICKER", new=mock.Mock(compile_to_file=mock.Mock(return_value="file"))
        ) as route:
            assert model.make_url() == "file"

        route.compile_to_file.assert_called_once_with(
            urls.CDN_URL, sticker_id=123, file_format="LOTTIE", size=4096, lossless=True
        )

    def test_make_url_uses_MEDIA_PROXY_when_not_LOTTIE(self, model):
        model.format_type = stickers.StickerFormatType.GIF

        with mock.patch.object(
            routes, "CDN_STICKER", new=mock.Mock(compile_to_file=mock.Mock(return_value="file"))
        ) as route:
            assert model.make_url() == "file"

        route.compile_to_file.assert_called_once_with(
            urls.MEDIA_PROXY_URL, sticker_id=123, file_format="GIF", size=4096, lossless=True
        )

    def test_make_url_raises_TypeError_when_GIF_sticker_requested_as_APNG(self, model):
        model.format_type = stickers.StickerFormatType.GIF

        with pytest.raises(TypeError):
            model.make_url(file_format="APNG")

    def test_make_url_raises_TypeError_when_APNG_sticker_requested_as_AWEBP_or_GIF(self, model):
        model.format_type = stickers.StickerFormatType.APNG

        with pytest.raises(TypeError):
            model.make_url(file_format="AWEBP")

        with pytest.raises(TypeError):
            model.make_url(file_format="GIF")

    def test_make_url_raises_TypeError_when_PNG_sticker_requested_as_animated_format(self, model):
        model.format_type = stickers.StickerFormatType.PNG

        with pytest.raises(TypeError):
            model.make_url(file_format="APNG")

        with pytest.raises(TypeError):
            model.make_url(file_format="AWEBP")

        with pytest.raises(TypeError):
            model.make_url(file_format="GIF")

    def test_make_url_raises_TypeError_when_LOTTIE_sticker_requested_as_non_LOTTIE_format(self, model):
        model.format_type = stickers.StickerFormatType.LOTTIE

        with pytest.raises(TypeError):
            model.make_url(file_format="PNG")

    def test_make_url_raises_TypeError_when_non_LOTTIE_sticker_requested_as_LOTTIE(self, model):
        model.format_type = stickers.StickerFormatType.PNG

        with pytest.raises(TypeError):
            model.make_url(file_format="LOTTIE")

    def test_make_url_applies_correct_settings_for_APNG(self, model):
        model.format_type = stickers.StickerFormatType.APNG

        with mock.patch.object(
            routes, "CDN_STICKER", new=mock.Mock(compile_to_file=mock.Mock(return_value="file"))
        ) as route:
            assert model.make_url(file_format="PNG") == "file"

        route.compile_to_file.assert_called_once_with(
            urls.MEDIA_PROXY_URL, sticker_id=123, file_format="PNG", size=4096, lossless=True
        )

    def test_make_url_applies_correct_settings_for_AWEBP(self, model):
        model.format_type = stickers.StickerFormatType.GIF

        with mock.patch.object(
            routes, "CDN_STICKER", new=mock.Mock(compile_to_file=mock.Mock(return_value="file"))
        ) as route:
            assert model.make_url(file_format="AWEBP") == "file"

        route.compile_to_file.assert_called_once_with(
            urls.MEDIA_PROXY_URL, sticker_id=123, file_format="AWEBP", size=4096, lossless=True
        )

    def test_make_url_applies_correct_settings_for_WEBP_lossless(self, model):
        model.format_type = stickers.StickerFormatType.PNG

        with mock.patch.object(
            routes, "CDN_STICKER", new=mock.Mock(compile_to_file=mock.Mock(return_value="file"))
        ) as route:
            assert model.make_url(file_format="WEBP", lossless=True) == "file"

        route.compile_to_file.assert_called_once_with(
            urls.MEDIA_PROXY_URL, sticker_id=123, file_format="WEBP", size=4096, lossless=True
        )

    def test_make_url_applies_correct_settings_for_WEBP_lossy(self, model):
        model.format_type = stickers.StickerFormatType.PNG

        with mock.patch.object(
            routes, "CDN_STICKER", new=mock.Mock(compile_to_file=mock.Mock(return_value="file"))
        ) as route:
            assert model.make_url(file_format="WEBP", lossless=False) == "file"

        route.compile_to_file.assert_called_once_with(
            urls.MEDIA_PROXY_URL, sticker_id=123, file_format="WEBP", size=4096, lossless=False
        )

    def test_make_url_applies_no_extra_settings_for_non_special_formats(self, model):
        model.format_type = stickers.StickerFormatType.PNG

        with mock.patch.object(
            routes, "CDN_STICKER", new=mock.Mock(compile_to_file=mock.Mock(return_value="file"))
        ) as route:
            assert model.make_url(file_format="JPEG") == "file"

        route.compile_to_file.assert_called_once_with(
            urls.MEDIA_PROXY_URL, sticker_id=123, file_format="JPEG", size=4096, lossless=True
        )



# Copyright (c) 2020 Nekokatt
# Copyright (c) 2021-present davfsa
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import annotations

import pathlib

from pipelines import config
from pipelines import nox

STUBGEN_GENERATE = [
    "hikari/__init__.py",
    "hikari/api/__init__.py",
    "hikari/events/__init__.py",
    "hikari/impl/__init__.py",
    "hikari/interactions/__init__.py",
]


@nox.session(requires=["generate-stubs"])
def mypy(session: nox.Session) -> None:
    """Perform static type analysis on Python source code using mypy."""
    nox.sync(session, self=True, extras=["speedups", "server"], groups=["mypy"])

    session.run("mypy", "-p", config.MAIN_PACKAGE, "--config", config.PYPROJECT_TOML)
    session.run("mypy", "-p", config.EXAMPLE_SCRIPTS, "--config", config.PYPROJECT_TOML)


@nox.session()
def generate_stubs(session: nox.Session) -> None:
    """Generate the stubs for the package."""
    nox.sync(session, groups=["mypy", "ruff"])

    session.run("stubgen", *STUBGEN_GENERATE, "-o", ".", "--include-private", "--no-import")

    stub_paths = [path + "i" for path in STUBGEN_GENERATE]

    # At the time of writing, sorting imports is not done when running formatting
    # and needs to be done with ruff check
    # see: https://docs.astral.sh/ruff/formatter/#sorting-imports
    session.run("ruff", "format", *stub_paths)
    session.run("ruff", "check", "--select", "I", "--fix", *stub_paths)

    for stub_path in stub_paths:
        path = pathlib.Path(stub_path)

        content = path.read_bytes()

        path.write_text(
            "# DO NOT MANUALLY EDIT THIS FILE!\n"
            "# This file was automatically generated by `nox -s generate-stubs`\n\n" + content.decode()
        )



# Copyright (c) 2020 Nekokatt
# Copyright (c) 2021-present davfsa
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import annotations

import mock
import pytest

from hikari import snowflakes
from hikari import stage_instances


@pytest.fixture
def mock_app():
    return mock.Mock()


class TestStageInstance:
    @pytest.fixture
    def stage_instance(self, mock_app):
        return stage_instances.StageInstance(
            app=mock_app,
            id=snowflakes.Snowflake(123),
            channel_id=snowflakes.Snowflake(6969),
            guild_id=snowflakes.Snowflake(420),
            topic="beanos",
            privacy_level=stage_instances.StageInstancePrivacyLevel.GUILD_ONLY,
            discoverable_disabled=True,
            scheduled_event_id=snowflakes.Snowflake(1337),
        )

    def test_id_property(self, stage_instance):
        assert stage_instance.id == 123

    def test_app_property(self, stage_instance, mock_app):
        assert stage_instance.app is mock_app

    def test_channel_id_property(self, stage_instance):
        assert stage_instance.channel_id == 6969

    def test_guild_id_property(self, stage_instance):
        assert stage_instance.guild_id == 420

    def test_topic_property(self, stage_instance):
        assert stage_instance.topic == "beanos"

    def test_privacy_level_property(self, stage_instance):
        assert stage_instance.privacy_level == stage_instances.StageInstancePrivacyLevel.GUILD_ONLY

    def test_discoverable_disabled_property(self, stage_instance):
        assert stage_instance.discoverable_disabled is True

    def test_guild_scheduled_event_id_property(self, stage_instance):
        assert stage_instance.scheduled_event_id == 1337
